@@ -5,11 +5,8 @@ package Gtk::GLArea;
 require Gtk;
 require Exporter;
 require DynaLoader;
-require AutoLoader;
 
-use Carp;
-
-$VERSION = '0.7005';
+$VERSION = '0.7006';
 
 @ISA = qw(Exporter DynaLoader);
 # Items to export into callers namespace by default. Note: do not export
