@@ -787,23 +787,30 @@ print "package $opt{FilePrefix}::Types;\n";
 foreach (sort keys %object) {
 	if (defined $object{$_}->{parent}) {
 		my ($pp) = $object{$object{$_}->{parent}}->{perlname} || perlize($object{$_}->{parent});
-		if (!$Lazy) {
-			print "\@$object{$_}->{perlname}::ISA = '$pp';\n";
-		} else {
-			print "\@$object{$_}->{perlname}::ISA = ('Gtk::_LazyLoader');\n";
-			print "\@$object{$_}->{perlname}::_ISA = '$pp';\n";
-		}
+		print "\@$object{$_}->{perlname}::ISA = '$pp';\n";
 	}
 }
 
-if ($Lazy) {
-	print "\n\nGtk::Object::_bootstrap('Gtk::Object');\n\n";
-}
-
 print "1;\n";
 
+if ($Lazy) {
+open(OUT,">build/$opt{FilePrefix}TypesLazy.pm") or die "Unable to write to build/$opt{FilePrefix}TypesLazy.pm: $!";
+print "\n\n# Do not edit this file, as it is automatically generated by gendefs.pl\n\n";
 
 
+print "package $opt{FilePrefix}::Types;\n";
+foreach (sort keys %object) {
+	if (defined $object{$_}->{parent}) {
+		my ($pp) = $object{$object{$_}->{parent}}->{perlname} || perlize($object{$_}->{parent});
+		print "\@$object{$_}->{perlname}::_ISA = '$pp';\n";
+		print "\@$object{$_}->{perlname}::ISA = 'Gtk::_LazyLoader';\n";
+	}
+}
+
+print "1;\n";
+
+}
+
 open(OUT,">build/$opt{FilePrefix}Defs.c")  or die "Unable to write to build/$opt{FilePrefix}Defs.c: $!";
 
 print <<"EOT";
@@ -1333,40 +1340,22 @@ DESTROY(self)
 EOT
 }
 
-open(OUT,">build/objects.xsh") or die "Unable to write to objects.xsh: $!";
-
-print "\n\n# Do not edit this file, as it is automatically generated by gendefs.pl\n\n";
-
 if ($Lazy) {
-
-foreach (sort keys %object) {
-	next if not length $object{$_}->{xsname};
-	print <<"EOT";
-
-MODULE = $Module	PACKAGE = $object{$_}->{perlname}		PREFIX = $object{$_}->{prefix}_
-
-void
-$object{$_}->{prefix}__bootstrap(self)
-	CODE:
-	{
-	#ifdef $object{$_}->{cast}
-		extern void boot_$object{$_}->{xsname}(CV *cv);
-		callXS (boot_$object{$_}->{xsname}, cv, mark);
-	#endif
-	}
-
-EOT
+	open(OUT,">build/$opt{FilePrefix}objects.xsh") or die "Unable to write to $opt{FilePrefix}objects.xsh: $!";
+	print "\n\n/* Do not edit this file, as it is automatically generated by gendefs.pl*/\n\n";
+} else {
+	open(OUT,">build/objects.xsh") or die "Unable to write to objects.xsh: $!";
+	print "\n\n# Do not edit this file, as it is automatically generated by gendefs.pl\n\n";
 }
 
-} else {
 
-print "MODULE = $Module	PACKAGE = $Module\n\n";
+print "MODULE = $Module	PACKAGE = $Module\n\n" unless $Lazy;
 
 foreach (sort keys %object) {
 	next if not length $object{$_}->{xsname};
 	next if $object{$_}->{perlname} eq $Module;
+	print "BOOT:\n" unless $Lazy;
 	print <<"EOT";
-BOOT:
 {
 	#ifdef $object{$_}->{cast}
                 extern void boot_$object{$_}->{xsname}(CV *cv);
@@ -1377,8 +1366,6 @@ BOOT:
 EOT
 }
 
-}
-
 close(OUT);
 select(STDOUT);
 
